unit main;

{
Graphical Hints - Image in Virtual Tree View Node Hint (Extending TVirtualTreeHintWindow)
http://zarko-gajic.iz.hr/graphical-hints-image-in-virtual-treeview-node-hint-extending-tvirtualtreehintwindow

Requires:

1) Virtual Treeview: http://www.delphi-gems.com/index.php/controls/virtual-treeview
2) Debenu PDF Library: http://www.debenu.com/products/development/debenu-pdf-library/

by arko Gaji
}


interface

uses
  Vcl.FileCtrl, System.Types, System.IOUtils, Vcl.Imaging.jpeg, DebenuPDFLibrary0916,
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, VirtualTrees,
  Vcl.ExtCtrls;

type
  TTreeNodeData = record
  private
    fPreview : TJpegImage;
    function GetPreview: TJpegImage;
  public
    Name: string;
    RelativeName : string;
    IsFolder : boolean;
    property Preview : TJpegImage read GetPreview;
  end;
  PTreeNodeData = ^TTreeNodeData;

  TVirtualStringTree = class(VirtualTrees.TVirtualStringTree)
  private
    fHintModeImage: boolean;
    procedure SetHintModeImage(const Value: boolean);
  public
    constructor Create(AOwner: TComponent); override;
  public
    function GetHintWindowClass: THintWindowClass; override;
    property HintModeImage : boolean read fHintModeImage write SetHintModeImage;
  end;

  TVirtualTreeHintWindowEx = class(TVirtualTreeHintWindow)
    function CalcHintRect(MaxWidth: Integer; const AHint: string; AData: TCustomData): TRect; override;
    procedure Paint; override;
  end;



  TMainForm = class(TForm)
    pnlTop: TPanel;
    edRootDirectory: TEdit;
    btnProcessFolder: TButton;
    pnlTree: TPanel;
    tree: TVirtualStringTree;
    pnlTreeTop: TPanel;
    chkNodeImageHint: TCheckBox;
    procedure FormCreate(Sender: TObject);
    procedure treeFreeNode(Sender: TBaseVirtualTree; Node: PVirtualNode);
    procedure treeGetImageIndex(Sender: TBaseVirtualTree;
      Node: PVirtualNode; Kind: TVTImageKind; Column: TColumnIndex;
      var Ghosted: Boolean; var ImageIndex: Integer);
    procedure treeGetText(Sender: TBaseVirtualTree; Node: PVirtualNode;
      Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);
    procedure treePaintText(Sender: TBaseVirtualTree;
      const TargetCanvas: TCanvas; Node: PVirtualNode;
      Column: TColumnIndex; TextType: TVSTTextType);
    procedure treeGetHint(Sender: TBaseVirtualTree; Node: PVirtualNode;
      Column: TColumnIndex; var LineBreakStyle: TVTTooltipLineBreakStyle;
      var HintText: string);
    procedure chkNodeImageHintClick(Sender: TObject);
    procedure btnProcessFolderClick(Sender: TObject);
  private
    fRootDirectory: string;
    procedure SetRootDirectory(const Value: string);
    property RootDirectory : string read fRootDirectory write SetRootDirectory;

    procedure FillTree(const extension : string = '.pdf');
  public
    { Public declarations }
  end;




var
  MainForm: TMainForm;

implementation

{$R *.dfm}

procedure TMainForm.btnProcessFolderClick(Sender: TObject);
var
  selectedRoot : string;
begin
  if NOT SelectDirectory('Select root directory', '', selectedRoot) then Exit;

  RootDirectory := selectedRoot;
end;

procedure TMainForm.chkNodeImageHintClick(Sender: TObject);
begin
  tree.HintModeImage := chkNodeImageHint.Checked;
end;

procedure TMainForm.FillTree(const extension : string = '.pdf');
  procedure ProcessDirectory(const dir : string; const parentNode : PVirtualNode);
  var
    dNode, fNode : PVirtualNode;
    subFolders, subFiles : TStringDynArray;
    subFolder, subFile : string;
    treeNodeData : PTreeNodeData;
  begin
    subFiles := TDirectory.GetFiles(
                    dir,
                    TSearchOption.soTopDirectoryOnly,
                    function(const Path: string; const SearchRec: TSearchRec): Boolean
                    begin
                      result := SameText(ExtractFileExt(SearchRec.Name), extension);
                    end);

    for subFile in subFiles do
    begin
      fNode := tree.AddChild(parentNode);

      treeNodeData := PTreeNodeData(tree.GetNodeData(fNode));
      treeNodeData.Name := subFile;
      treeNodeData.RelativeName := Copy(subFile, 1 + Length(RootDirectory));
      treeNodeData.IsFolder := false;
    end;

    subFolders := TDirectory.GetDirectories(dir, TSearchOption.soTopDirectoryOnly, nil);

    for subFolder in subFolders do
    begin
      dNode := tree.AddChild(parentNode);

      treeNodeData := PTreeNodeData(tree.GetNodeData(dNode));
      treeNodeData.Name := subFolder;
      treeNodeData.RelativeName := Copy(subFolder, 1 + Length(RootDirectory));
      treeNodeData.IsFolder := true;

      ProcessDirectory(subFolder, dNode);
    end;
  end; (*ProcessDirectory*)
begin

  tree.Clear;

  tree.BeginUpdate;
  try
    ProcessDirectory(RootDirectory, tree.RootNode);
    tree.ValidateNode(nil, True);
    tree.FullExpand();
  finally
    tree.EndUpdate;
  end;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  tree.NodeDataSize := SizeOf(TTreeNodeData);

  //ensure node hints specific are displayed
  tree.HintMode := hmHint;
  tree.ShowHint := true;

  //use image hint mode
  chkNodeImageHint.Checked := true;

  {$IFDEF DEBUG}
  RootDirectory := 'D:\PDF\Documents';
  {$ENDIF}
end;

procedure TMainForm.SetRootDirectory(const Value: string);
begin
  fRootDirectory := Value;

  edRootDirectory.Text := RootDirectory;

  FillTree;
end;

procedure TMainForm.treeFreeNode(Sender: TBaseVirtualTree; Node: PVirtualNode);
var
  data: PTreeNodeData;
begin
  data := Sender.GetNodeData(Node);
  if Assigned(data.fPreview) then data.fPreview.Free;
  Finalize(data^);
end;

procedure TMainForm.treeGetHint(Sender: TBaseVirtualTree;
  Node: PVirtualNode; Column: TColumnIndex;
  var LineBreakStyle: TVTTooltipLineBreakStyle; var HintText: string);
var
  data: PTreeNodeData;
begin
  data := Sender.GetNodeData(Node);

  HintText := data.RelativeName;
end;

procedure TMainForm.treeGetImageIndex(Sender: TBaseVirtualTree;
  Node: PVirtualNode; Kind: TVTImageKind; Column: TColumnIndex;
  var Ghosted: Boolean; var ImageIndex: Integer);
var
  data: PTreeNodeData;
begin
  data := Sender.GetNodeData(Node);

  if (Column = 0) AND (Kind in [ikNormal,ikSelected]) then
  begin
    if data.IsFolder then
      ImageIndex := 0
    else
      ImageIndex := 1;
  end;
end;

procedure TMainForm.treeGetText(Sender: TBaseVirtualTree; Node: PVirtualNode;
  Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);
var
  Data: PTreeNodeData;
begin
  Data := Sender.GetNodeData(Node);

  if data.IsFolder then
    CellText := data.RelativeName
  else
    CellText := ExtractFileName(data.Name);
end;

procedure TMainForm.treePaintText(Sender: TBaseVirtualTree;
  const TargetCanvas: TCanvas; Node: PVirtualNode; Column: TColumnIndex;
  TextType: TVSTTextType);
var
  data: PTreeNodeData;
begin
  data := Sender.GetNodeData(Node);

  if (Column = 0) then
  begin
    if data.IsFolder then TargetCanvas.Font.Style := [fsBold];
  end;
end;

{ TVirtualStringTree }

constructor TVirtualStringTree.Create(AOwner: TComponent);
begin
  fHintModeImage := false;
  inherited;
end;

function TVirtualStringTree.GetHintWindowClass: THintWindowClass;
begin
  if HintModeImage then
    result := TVirtualTreeHintWindowEx
  else
    result := inherited;
end;

procedure TVirtualStringTree.SetHintModeImage(const Value: boolean);
begin
  fHintModeImage := Value;

  if fHintModeImage then
    HintAnimation := hatNone //so Paint procedure is called and can be overriden
  else
    HintAnimation := hatSystemDefault; //default
end;

{ TVirtualTreeHintWindowEx }

function TVirtualTreeHintWindowEx.CalcHintRect(MaxWidth: Integer;
  const AHint: string; AData: TCustomData): TRect;
var
  r : TRect;
  hd : TVTHintData;
  jpg : TJpegImage;
  nodeData : PTreeNodeData;
begin
  r := inherited;

  hd := PVTHintData(AData)^;

  if Assigned(hd.Tree) AND Assigned(hd.Node) AND (HintData.Tree is TVirtualStringTree) AND (TVirtualStringTree(HintData.Tree).HintModeImage) then
  begin
    nodeData := HintData.Tree.GetNodeData(hd.Node);

    jpg := nodeData.Preview;

    if Assigned(jpg) AND (NOT jpg.Empty) then
    begin
      r.Height := r.Height + jpg.Height + 4;
      if r.Width < jpg.Width + 4 then r.Width := jpg.Width + 4;
    end;
  end;

  result :=  r;
end;

procedure TVirtualTreeHintWindowEx.Paint; //called only if HintAnimation = hatNonde
var
  jpg : TJpegImage;
  nodeData : PTreeNodeData;
begin
  InternalPaint(0, 0);

  if Assigned(HintData.Tree) AND Assigned(HintData.Node) AND (HintData.Tree is TVirtualStringTree) AND (TVirtualStringTree(HintData.Tree).HintModeImage) then
  begin
    nodeData := HintData.Tree.GetNodeData(HintData.Node);

    jpg := nodeData.Preview;

    if Assigned(jpg) AND (NOT jpg.Empty) then Canvas.Draw(2, -2 + Height - jpg.Height, jpg);
  end;
end;


{ TTreeNodeData }

function TTreeNodeData.GetPreview: TJpegImage;
var
  fh, pr : integer;
  ms : TMemoryStream;
  bmp : TBitmap;
  pdfFileName : string;
begin
  if fPreview = nil then fPreview := TJpegImage.Create;

  if (NOT IsFolder) AND fPreview.Empty then
  begin
    pdfFileName := self.Name;

    //http://www.debenu.com/products/development/debenu-pdf-library/
    with TQuickPDF.Create do
    try
      fh := DAOpenFileReadOnly(pdfFileName, '');
      if fh <> 0 then
      try
        pr := DAFindPage(fh, 1);
        if pr <> 0 then
        begin
          ms := TMemoryStream.Create;
          try
            DARenderPageToStream(fh, pr, 1, 24, ms);
            ms.Seek(0,0);

            fPreview.LoadFromStream(ms);
          finally
            ms.Free;
          end;
        end;
      finally
        DACloseFile(fh);
      end;
    finally
      Free;
    end;
  end;

  result := fPreview;
end;

end.
